/*
 * Copyright 2015 Tom Schons - TU Darmstadt, Germany
 * Released under GPLv3. See LICENSE.txt for details.
 */
package movement;

import core.*;
import movement.naturaldisaster.*;
import movement.utils.Group;

/**
 * This movement model simulates disaster relief efforts after a natural
 * disaster, such as, for example, typhoons, floodings, earthquakes, etc.
 * <p>
 * The model covers relief efforts from day zero until the end of a defined
 * period (specifiable via the settings file) after the disaster occurred. This
 * class is the main movement model class. Further sub-models are listed below
 * and put into separate classes for convenience and modularity reasons.
 *
 * @author Tom Schons
 * @author Adam Walker
 */

public class NaturalDisasterUKMovementModel extends ExtendedMovementModel
        implements Group {

    /*
     * Constants for importing settings from default settings file
     */
    /**
     * Role of an individual
     */
    public static final String ROLE = "role";
    /**
     * Number of days for the simulation
     */
    public static final String NUMBER_OF_DAYS = "nbrOfDays";

    /**
     * Length of the day in seconds
     */
    private static final int SECONDS_IN_A_DAY = 24 * 60 * 60;

    /*
     * Sub-movement models we load for scenario creation
     */
    private SleepActivityMovement goSleepMM;
    private ArrivalActivityMovement airportArrivalMM;
    private GoToAirportActivityMovement goToAirportMM;
    private NonInjuredPopulationActivityMovement nonInjuredMM;
    private InjuredPopulationActivityMovement injuredMM;
    private DisasterReliefActivityMovement disasterReliefMM;
    private SearchAndRescueActivityMovement searchRescueMM;

    /*
     * Different modes in which an individual can be in (depends on role)
     */
    /**
     * Individual is in sleep mode (during evening and night)
     */
    private static final int SLEEP_MODE = 10;
    /**
     * Individual just arrived in the disaster area (in our case via the
     * airport)
     */
    private static final int ARRIVAL_MODE = 11;
    /**
     * Individual is in disaster relief mode (used for simulating disaster
     * relief
     * organizations)
     */
    private static final int DISASTER_RELIEF_MODE = 12;
    /**
     * Individual is going to the airport & leaving the disaster area
     */
    private static final int GO_AIRPORT_MODE = 14;
    /**
     * Individual is in search and rescue mode (e.g. mostly urban search and
     * rescue
     * teams)
     */
    private static final int SEARCH_RESCUE_MODE = 16;
    /**
     * Individual is in injured mode
     */
    private static final int INJURED_MODE = 17;
    /**
     * Individual is in non-injured mode
     */
    private static final int HEALTHY_MODE = 18;

    /*
     * Different roles an individual can have
     */
    /**
     * Injured local population role
     */
    public static final String INJURED_ROLE = "Injured";
    /**
     * Healthy local population role
     */
    public static final String HEALTHY_ROLE = "Healthy";
    /**
     * Search-and-Rescue team role
     */
    public static final String SEARCH_RESCUE_ROLE = "SnR";
    /**
     * Disaster relief organization role
     */
    public static final String DRO_ROLE = "DRO";

    /**
     * Current mode in which the individual currently is in
     */
    private int mode;

    /**
     * The role of the individual that this model is run on
     */
    private final String role;

    /**
     * Number of days
     */
    private int nbrOfDays;

    /**
     * The current day
     */
    private int dayCounter;

    /**
     * Creates a new instance of NaturalDisasterMovementModel
     *
     * @param settings
     */
    public NaturalDisasterUKMovementModel(Settings settings) {
        super(settings);
        goSleepMM = new SleepActivityMovement(settings);
        airportArrivalMM = new ArrivalActivityMovement(settings);
        goToAirportMM = new GoToAirportActivityMovement(settings);
        injuredMM = new InjuredPopulationActivityMovement(settings);
        nonInjuredMM = new NonInjuredPopulationActivityMovement(settings);
        searchRescueMM = new SearchAndRescueActivityMovement(settings);
        disasterReliefMM = new DisasterReliefActivityMovement(settings);

        this.dayCounter = 0;

        if (settings.contains(ROLE)) {
            role = settings.getSetting(ROLE);
        } else {
            throw new SettingsError(
                    "No Role specified (you need to specify Group.role when " +
                            "using NaturalDisasterMovementModel)");
        }
        if (settings.contains(NUMBER_OF_DAYS)) {
            nbrOfDays = settings.getInt(NUMBER_OF_DAYS);
        } else {
            throw new SettingsError(
                    "No value for the number of days " + "specified");
        }

        setInitialActivity();
    }

    /**
     * Creates a new instance of NaturalDisasterMovementModel from a prototype
     *
     * @param prototype
     */
    public NaturalDisasterUKMovementModel(NaturalDisasterUKMovementModel
                                                  prototype) {
        // Import settings with all parameters from prototype
        super(prototype);
        goSleepMM = new SleepActivityMovement(prototype.goSleepMM);
        airportArrivalMM = new ArrivalActivityMovement(prototype
                .airportArrivalMM);
        goToAirportMM = new GoToAirportActivityMovement(prototype
                .goToAirportMM);
        injuredMM = new InjuredPopulationActivityMovement(prototype.injuredMM);
        nonInjuredMM = new NonInjuredPopulationActivityMovement(prototype
                .nonInjuredMM);
        searchRescueMM = new SearchAndRescueActivityMovement(prototype
                .searchRescueMM);
        disasterReliefMM = new DisasterReliefActivityMovement(prototype
                .disasterReliefMM);

        this.dayCounter = 0;

        this.role = new String(prototype.getRole());
        this.nbrOfDays = prototype.getNbrOfDays();

        setInitialActivity();
    }

    private void setInitialActivity() {
        // Current movement model (e.g. activity) for the specific groups
        // are set in
        // accordance with their role
        switch (role) {
            case INJURED_ROLE: {
                // Injured individuals obviously are already within the
                // disaster area, thus they
                // start their activity in sleep mode
                setCurrentMovementModel(goSleepMM);
                mode = SLEEP_MODE;
                break;
            }
            case HEALTHY_ROLE: {
                // Non-injured individuals obviously are already within the
                // disaster area, thus
                // they start their activity in sleep mode
                setCurrentMovementModel(goSleepMM);
                mode = SLEEP_MODE;
                break;
            }
            case SEARCH_RESCUE_ROLE: {
                // Search and rescue groups usually arrive via the airport
                // after the disaster
                // occurred, thus they start their activity in the
                // airportArrivalMM movement
                // model
                setCurrentMovementModel(airportArrivalMM);
                mode = ARRIVAL_MODE;
                break;
            }
            case DRO_ROLE: {
                // Disaster relief organizations are usually not present
                // within the disaster
                // area and thus start their activity in the
                // airportArrivalMM movement model
                setCurrentMovementModel(airportArrivalMM);
                mode = ARRIVAL_MODE;
                break;
            }
            default:
                throw new SettingsError(
                        "Invalid role set (was '" + role + "')");
        }
    }

    /**
     * Method is called between each getPath() request when the current MM
     * is ready
     * (isReady() method returns true). Subclasses implement all changes of
     * state
     * that need to be made.
     * <p>
     * This functions holds all basic structures for switching the different
     * activities in a disater area
     *
     * @return true if success
     */
    @Override
    public boolean newOrders() {
        // Always calculating the number of the days passed by when calling
        // newOrders()
        if ((SimClock.getTime() >
                ((this.dayCounter + 1) * SECONDS_IN_A_DAY)) &&
                (!(this.dayCounter == this.nbrOfDays))) {
            // Simulation is not yet finished but another day passed by
            this.dayCounter++;
            // Update the sleep activity since counter is usually updated
            // after we finished
            // the sleep activity -> need to be updated retroactively
            goSleepMM.updateDayCounter(this.dayCounter);
        }

        // Now checking what mode we are operating in and then proceeding to
        // do what is
        // necessary according to that mode of operation
        switch (mode) {
            // Switching between the different modes of activity
            case SLEEP_MODE: {
                if (goSleepMM.isReady()) {
                    // If true we are done with sleeping and want to become
                    // active again since the
                    // next day has just started
                    if (this.dayCounter == this.nbrOfDays) {
                        // If we end up here the simulation is finished ->
                        // we idle forever
                        break;
                    }
                    if (role.equals(INJURED_ROLE)) {
                        // Switching to the next movement model
                        setCurrentMovementModel(injuredMM);
                        // set mode to accurate model
                        mode = INJURED_MODE;
                        // Activate the correct movement model again and
                        // update its dayCounter
                        injuredMM.Activate(this.dayCounter);
                    }
                    if (role.equals(HEALTHY_ROLE)) {
                        // Switching to the next movement model
                        setCurrentMovementModel(nonInjuredMM);
                        // set mode to accurate model
                        mode = HEALTHY_MODE;
                        // Activate the correct movement model again and
                        // update its dayCounter
                        nonInjuredMM.Activate(this.dayCounter);
                    }
                    if (role.equals(SEARCH_RESCUE_ROLE)) {
                        // Switching to the next movement model
                        if (!searchRescueMM.getGoToAirport()) {
                            // we remain in the search and rescue movement
                            // model
                            setCurrentMovementModel(searchRescueMM);
                            // set mode to accurate model
                            mode = SEARCH_RESCUE_MODE;
                            // Activate the correct movement model again and
                            // update its dayCounter
                            searchRescueMM.Activate(this.dayCounter);
                        } else {
                            // we leave the search and rescue model for
                            // going to the airport
                            setCurrentMovementModel(goToAirportMM);
                            // set mode to accurate model
                            mode = GO_AIRPORT_MODE;
                            // Set the last location of the search and
                            // rescue activity to be the initial
                            // location of the go to airport activity
                            goToAirportMM.setInitialLocation(searchRescueMM
                                    .getLastLocation().clone());
                            // Activate the correct movement model again and
                            // update its dayCounter
                            goToAirportMM.Activate();
                        }
                    }
                    if (role.equals(DRO_ROLE)) {
                        // Switching to the next movement model
                        setCurrentMovementModel(disasterReliefMM);
                        // set mode to accurate model
                        mode = DISASTER_RELIEF_MODE;
                        // Activate the correct movement model again and
                        // update its dayCounter
                        disasterReliefMM.Activate(this.dayCounter);
                    }
                }
                break;
            }
            case ARRIVAL_MODE: {
                if (airportArrivalMM.isReady()) {
                    // If true we are done with the arrival at airport part
                    // and want to become
                    // active in the disaster area
                    if (role.equals(SEARCH_RESCUE_ROLE)) {
                        // Mandatory to set initial location in order to
                        // avoid null pointer exceptions
                        // since this activity starts after the arrival at
                        // airport activity
                        searchRescueMM.setInitialLocation(airportArrivalMM
                                .getLastLocation().clone());
                        setCurrentMovementModel(searchRescueMM);
                        mode = SEARCH_RESCUE_MODE;
                        // Activate the correct movement model again and
                        // update its dayCounter
                        searchRescueMM.Activate(this.dayCounter);
                    } else if (role.equals(DRO_ROLE)) {
                        // Mandatory to set initial location in order to
                        // avoid null pointer exceptions
                        // since this activity starts after the arrival at
                        // airport activity
                        disasterReliefMM.setInitialLocation(airportArrivalMM
                                .getLastLocation().clone());
                        setCurrentMovementModel(disasterReliefMM);
                        mode = DISASTER_RELIEF_MODE;
                        // Activate the correct movement model again and
                        // update its dayCounter
                        disasterReliefMM.Activate(this.dayCounter);
                    } else {
                        throw new SimError("Should not be in this mode");
                    }
                }
                break;
            }
            case DISASTER_RELIEF_MODE: {
                if (disasterReliefMM.isReady()) {
                    // If true we are done with the disaster relief activity
                    // for today and want to
                    // go home / to sleep (generally speaking)
                    if (role.equals(DRO_ROLE)) {
                        // Day is over, going home to sleep -> Activating
                        // sleep activity
                        setCurrentMovementModel(goSleepMM);
                        mode = SLEEP_MODE;
                        goSleepMM.Activate(this.dayCounter);
                    } else {
                        throw new SimError("Should not be in this mode");
                    }
                }
                break;
            }
            case GO_AIRPORT_MODE: {
                if (goToAirportMM.isReady()) {
                    // If true we are done with disaster relief and already
                    // at the airport -> so we
                    // solely idle here
                }
                break;
            }
            case SEARCH_RESCUE_MODE: {
                if (searchRescueMM.isReady()) {
                    // If true we are done with the search and rescue
                    // activity and want to go home /
                    // to sleep (generally speaking)
                    if (role.equals(SEARCH_RESCUE_ROLE)) {
                        // Day is over, going home to sleep -> Activating
                        // sleep activity
                        setCurrentMovementModel(goSleepMM);
                        mode = SLEEP_MODE;
                        goSleepMM.Activate(this.dayCounter);
                    } else {
                        throw new SimError("Should not be in this mode");
                    }
                }
                break;
            }
            case INJURED_MODE: {
                if (injuredMM.isReady()) {
                    // If true we are done with our daily activity and want
                    // to go home / to sleep
                    // (generally speaking)
                    if (role.equals(INJURED_ROLE)) {
                        // Day is over, going home to sleep -> Activating
                        // sleep activity
                        setCurrentMovementModel(goSleepMM);
                        mode = SLEEP_MODE;
                        goSleepMM.Activate(this.dayCounter);
                    } else {
                        throw new SimError("Should not be in this mode");
                    }
                }
                break;
            }
            case HEALTHY_MODE: {
                if (nonInjuredMM.isReady()) {
                    // If true we are done with our daily activity and want
                    // to go home / to sleep
                    // (generally speaking)
                    if (role.equals(HEALTHY_ROLE)) {
                        // Day is over, going home to sleep -> Activating
                        // sleep activity
                        setCurrentMovementModel(goSleepMM);
                        mode = SLEEP_MODE;
                        goSleepMM.Activate(this.dayCounter);
                    } else {
                        throw new SimError("Should not be in this mode");
                    }
                }
                break;
            }
            default:
                throw new SimError("Was in invalid mode " + mode);
        }
        return true;
    }

    @Override
    public Coord getInitialLocation() {
        // Returning the last location of each sub-movement model as the
        // initial
        // location since each sub-movement model is configured with it's
        // last location
        // being the initial location
        switch (role) {
            case SEARCH_RESCUE_ROLE: {
                return searchRescueMM.getLastLocation();
            }
            case DRO_ROLE: {
                return disasterReliefMM.getLastLocation();
            }
            case INJURED_ROLE: {
                return injuredMM.getLastLocation();
            }
            case HEALTHY_ROLE: {
                return nonInjuredMM.getLastLocation();
            }
            default: {
                throw new SettingsError(
                        "Could not get initial location for " + "role " +
                                role);
            }
        }
    }

    @Override
    public MovementModel replicate() {
        return new NaturalDisasterUKMovementModel(this);
    }

    /**
     * Fetch the role this node has.
     *
     * @return this hosts role
     */
    public String getRole() {
        return this.role;
    }

    public int getNbrOfDays() {
        return this.nbrOfDays;
    }

    @Override
    public int getGroup() {
        switch (this.role) {
            case INJURED_ROLE:
                return 1;
            case HEALTHY_ROLE:
            case SEARCH_RESCUE_ROLE:
            case DRO_ROLE:
            default:
                return 0;
        }
    }
}