package movement;

import core.Coord;
import core.Settings;
import core.SimScenario;
import input.WKTReader;
import movement.map.MapNode;
import movement.map.SimMap;
import movement.utils.Group;

import java.io.File;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;
import java.util.stream.Collectors;

/**
 * A class which models a node at a fixed location. This builds upon
 * StationaryMovement by permitting a file containing points on a map. Each
 * node is assigned in order to the point on the map indicated in the
 * supplied file.
 *
 * @author Adam Walker
 */
public class StationaryPointMapBasedMovement extends MapBasedMovement implements Group{

    private static final String LOCATIONS_FILES_SETTING = "locationsFiles";
    private static final String SNAP_POINTS_TO_MAP_S = "snapToMap";
    private static int assignedLocationCount = 0;
    private List<Coord> allLocations = new LinkedList<Coord>();
    private Coord location;

    /**
     * Creates a new instance based on a Settings object's configuration.
     *
     * @param settings object for reading settings info
     * @throws Throwable on fail to read settings
     */
    public StationaryPointMapBasedMovement(Settings settings) throws
            Throwable {
        super(settings);
        String[] locationsFiles;
        boolean snapToMap;

        // fetch the locations file(s) from the settings file
        locationsFiles = null;
        try {
            locationsFiles = settings.getCsvSetting(LOCATIONS_FILES_SETTING);
        } catch (Throwable t) {
            System.out.println(
                    "Reading the location files somehow failed - " + "" + "" +
                            "did you specify the correct path?");
            throw t;
        }

        // check if we want to snap the points to the map
        snapToMap = settings.getBoolean(SNAP_POINTS_TO_MAP_S, false);

        int nrofHosts = settings.getInt(SimScenario.NROF_HOSTS_S);
        try {
            for (String f : locationsFiles) {
                // read the coordinates from the WKT file
                this.allLocations.addAll(readPointsFromFile(f));
            }

            // warn the user if there are potential settings mistakes
            if (this.allLocations.size() == 0) {
                System.out.println(
                        "StationaryPointMapBasedMovement: You have 0 " +
                                "points");
            } else if (nrofHosts != this.allLocations.size()) {
                System.out.println(
                        "StationaryPointMapBasedMovement: You have " +
                                "different numbers of hosts and points");
                System.out.println("There are " + nrofHosts + " and " +
                        this.allLocations.size() + "points.");
            }

            // snap the coordinates to those on the map
            if (snapToMap) {
                // extract all node coordinates from the map
                List<Coord> mapCoords = this.getMap().getNodes().stream()
                        .map(MapNode::getLocation).collect(Collectors.toList
                                ());

                // find and set each point to it's closest map node
                for (int i = 0; i < this.allLocations.size(); i++) {
                    Coord p = this.allLocations.get(i);
                    double minDistance = Double.MAX_VALUE;

                    // search for closest point on map
                    for (int j = 0; j < mapCoords.size(); j++) {
                        Coord m = mapCoords.get(j);
                        double distance = p.distance(m);
                        // update if we've found a closer node
                        if (distance <= minDistance) {
                            minDistance = distance;
                            this.allLocations.set(i, m);
                        }
                    }
                }
            }

            // select the i'th location and increment the static counter
            this.location = this.allLocations.get
                    (StationaryPointMapBasedMovement
                            .assignedLocationCount++).clone();
            if (StationaryPointMapBasedMovement.assignedLocationCount >=
                    this.allLocations.size()) {
                StationaryPointMapBasedMovement.assignedLocationCount = 0;
            }
        } catch (Throwable t) {
            System.out.println("StationaryPointMapBasedMovement: reading the" +
                    " location file permanently failed");
            throw t;
        }
    }

    /**
     * Copy constructor.
     *
     * @param m MapBasedStationaryMovement prototype
     */
    public StationaryPointMapBasedMovement(StationaryPointMapBasedMovement m) {
        super(m);

        // notify the user if we are trying to set more than 1 node per
        // location
        if (StationaryPointMapBasedMovement.assignedLocationCount >=
                m.getLocationsCount()) {
            StationaryPointMapBasedMovement.assignedLocationCount = 0;
        }

        this.allLocations = m.allLocations;
        this.location = m.getLocation(StationaryPointMapBasedMovement
                .assignedLocationCount++).clone();
        if (StationaryPointMapBasedMovement.assignedLocationCount >=
                this.allLocations.size()) {
            StationaryPointMapBasedMovement.assignedLocationCount = 0;
        }
    }

    @Override
    public StationaryPointMapBasedMovement replicate() {
        return new StationaryPointMapBasedMovement(this);
    }

    /**
     * Returns the only location of this movement model
     *
     * @return the only location of this movement model
     */
    @Override
    public Coord getInitialLocation() {
        return this.location;
    }

    /**
     * Returns a single coordinate path (using the only possible coordinate)
     *
     * @return a single coordinate path
     */
    @Override
    public Path getPath() {
        Path p = new Path(0);
        p.addWaypoint(this.location);
        return p;
    }

    @Override
    public double nextPathAvailable() {
        return Double.MAX_VALUE; // no new paths available
    }

    /**
     * Fetch a location from the store.
     *
     * @param i the i'th location to get
     * @return the i'th location
     */
    public Coord getLocation(int i) {
        return this.allLocations.get(i);
    }

    /**
     * Get the number of locations we've read in from file.
     *
     * @return location count
     */
    public int getLocationsCount() {
        return this.allLocations.size();
    }

    /**
     * Fetch all the points stored.
     *
     * @return a list of all coordinates.
     */
    public List<Coord> getAllLocations() {
        return this.allLocations;
    }

    @Override
    public void setLocation(Coord location) {
        this.location = location;
    }

    /**
     * Reads a list of points from a file specified by name.
     *
     * @param file the file name of the points file
     * @return the list of points
     * @throws IOException on fail to read
     */
    private List<Coord> readPointsFromFile(String file) throws IOException {
        SimMap map = getMap();
        Coord offset = map.getOffset();
        List<Coord> locs = (new WKTReader()).readPoints(new File(file));

        for (Coord c : locs) {
            // mirroring all points if map data is mirrored
            if (map.isMirrored()) {
                c.setLocation(c.getX(), -c.getY());
            }
            c.translate(offset.getX(), offset.getY());
        }
        return locs;
    }

    @Override
    public int getGroup() {
        return 2;
    }
}
